require 'myredditnews/version'
require 'snoo'
require 'set'
require 'json'
require 'doodle'
require 'logger'
require 'feedzirra'
require 'nokogiri'
require 'pstore'

$logger = Logger.new(STDOUT)
# TODO get rid of the global variables
$delay = 4.5
$dry = false

module Myredditnews

  def self.delay; $delay; end
  def self.delay=d; $delay=d; end
  def self.dry; $dry; end
  def self.dry=d; $dry=d; end

  # TODO : remove the boiler plate below
  def self.with_dry(set_to)
    old = self.dry
    self.dry = set_to
    yield
    self.dry = old
  end
  def self.with_delay(set_to)
    old = self.delay
    self.delay = set_to
    yield
    self.delay = old
  end
  class RedditError < StandardError; end
  class Entry < Doodle
    has :author      , kind: String
    has :content     , kind: String
    has :url         , kind: String
    has :title       , kind: String
    has :date        , kind: String
    has :summary     , default: ''
    has :self_post   , default: true
    def submit(reddit, subreddit)
      opts = {}
      if self_post?
        opts[:kind] = 'self'
        opts[:text] = @content
      else
        opts[:kind] = 'link'
        opts[:url] = @url
      end
      $logger.info("Submitting #{opts[:kind]}: #{self}")
      puts "Dryness: #{Myredditnews.dry}"
      unless Myredditnews.dry
        response = reddit.submit(@title, subreddit, opts)
        $logger.info(response)
        throw_if_error response
        # TODO rate limiting should be smarter than this
        puts "Sleeping for #{Myredditnews.delay}"
        sleep Myredditnews.delay
      end
    end
    def to_s
      return "[#{truncate(@author, 10)}, #{truncate(@url, 20)}]"
    end
    def self_post?
      self_post
    end
    def to_json(*a)
      {
        "author"      => @author,
        "content"     => @content,
        "url"         => @url,
        "date"        => @date,
        "summary"     => @summary,
        "title"       => @title, }.to_json(*a)
    end
    private
    def truncate str, width
      str.size > width ? (str[0..width-1] + "...") : str
    end
    def throw_if_error response
      unless response['json']['errors'].empty?
        raise RedditError, response['json']['errors']
      end
    end
  end

  class Formatter < Doodle
    # @format is the "format" of the title only
    has :format, kind: String
    def fmt(entry)
      new_title = @format.gsub(/\{\{(.+?)\}\}/) { |match| entry.send($1) }
      new_entry = entry.clone
      new_entry.title = new_title
      new_entry.content = format_content(new_entry)
      if new_entry.content.empty?
        new_entry.self_post = false
      else
        new_entry.content += "\n\n" + md_link('source', new_entry.url)
      end
      return new_entry
    end
    def self.default
      self.new '{{title}} - {{author}}'
    end
    private 
    def md_link(text, url)
       "[#{text}](#{url})"
    end
    def format_content entry
      # maybe translate to markdown?
      return filter_html(entry.content) if !entry.content.empty?
      return filter_html(entry.summary) if !entry.summary.empty?
      return ''
    end
    def filter_html html
      Nokogiri::HTML(html).xpath("//text()").remove.to_s
    end
  end

  class FeedFetchingError < StandardError
    attr_reader :parent
    def initialize parent
      @parent = parent
    end
  end

  class Source < Doodle
    has :source, kind: String
    has :name, default: nil
    has :formatter, kind: Formatter, default: Formatter.default
    def safe_get field
      # not using sanitize because of some encoding bug.
      # field.nil? ? "" : field.sanitize
      field.nil? ? "" : field
    end
    def fetch
      feed = Feedzirra::Feed.fetch_and_parse(@source)
      s = self
      feed.entries.map do |feed_entry|
        Myredditnews::Entry.new do 
          author s.safe_get(feed_entry.author)
          content s.safe_get(feed_entry.content)
          title s.safe_get(feed_entry.title)
          summary s.safe_get(feed_entry.summary)
          url feed_entry.url
          date feed_entry.published.to_s
        end 
      end
    rescue => e
      raise FeedFetchingError, e
    end
    def format entry
      formatter.fmt(entry)
    end
  end

  class History
    def initialize path
      @path         = path
      @posted_queue = []
      @db           = PStore.new(File.expand_path @path)
    end
    def filter entries
      @db.transaction do
        entries.reject { |e| @db.root? e }
      end
    end
    # TODO : the only reason this method was not removed is because it's 
    # used in some tests
    def posted? entry
      return true if @posted_queue.include? entry
      @db.transaction do
        @db.root? entry
      end
    end
    def posted! entry
      @posted_queue << entry
    end
    def append_entries!
      return if Myredditnews.dry
      @db.transaction do
        @posted_queue.each do |e|
          @db[e] = 0
        end
      end
    end
    def delete_all!
      @db = nil
      File.delete @path if File.exists? @path
    end
  end

  class App < Doodle
    has :username         , kind: String
    has :password         , kind: String
    has :useragent        , default: "Myredditnews v#{Myredditnews::VERSION}"
    has :history          , kind: History do
      from String do |path|
        History.new(path)
      end
    end
    has :target_subreddit , kind: String
    has :sources, collect: Source
    def sync
      $logger.info("Mode: dry = #{Myredditnews.dry}")
      reddit = Snoo::Client.new(useragent: useragent)
      $logger.info("Logging in with #{@username}")
      reddit.log_in @username, @password
      total = 0
      fetch_sources().each do |source, submissions|
        only_new(submissions).each do |submission|
          begin
            source.format(submission).submit(reddit, @target_subreddit)
            @history.posted! submission
            total += 1
          rescue RedditError => e
            puts "Error submitting #{submission}"
            puts "Reason: #{e.message}"
          rescue => e
            puts "Failed submitting #{submission}"
            puts "Message: #{e.message}"
            puts e.backtrace
          end
        end
      end
      $logger.info("Sent #{total} entries to r/#{@target_subreddit}")
      reddit.log_out
      history.append_entries!
    end
    def only_new submissions
      @history.filter(submissions)
    end
    def fetch_sources
      sources = {}
      @sources.each do |source|
        begin
          sources[source] = source.fetch
        rescue FeedFetchingError => e
          $logger.info("Could not fetch feed #{source.source}. Reason:")
          $logger.info("Exception: #{e.message}. Backtrace:")
          $logger.info(e.parent.backtrace.drop(1).map {|s| "\t#{s}"}.join("\n"))
        end
      end
      sources
    end
  end
end
